#include "alloc.h"
#include "exception.h"
#include "gdt.h"
#include "idt.h"
#include "interrupt.h"
#include "io.h"
#include "irq.h"
#include "klibc.h"
#include "kthread.h"
#include "mem.h"
#include "multiboot.h"
#include "paging.h"
#include "pit.h"
#include "serial.h"
#include "stack.h"
#include "stdarg.h"
#ifdef RUN_TEST
#include "test.h"
#endif
#include "time.h"
#include "types.h"
#include "vga.h"

#define CHECK_FLAG(flags, bit) ((flags) & (1 << (bit)))

void idleThread(void *arg)
{
    (void)arg;
    while (1) {
        printIntDetails((jiffies / HZ), GREEN, BLACK, 0, VGA_HEIGHT - 1);
        kthreadYield();
    }
}

// Multiboot information available here :
// https://www.gnu.org/software/grub/manual/multiboot/multiboot.html#kernel_002ec
// https://www.gnu.org/software/grub/manual/multiboot/html_node/Boot-information-format.html#Boot%20information%20format
void kmain(unsigned long magic, unsigned long addr)
{
    unsigned long upper_mem = 0;
    VGASetup(BLACK, GREEN);
    cursorEnable(14, 15);

    printf("Setting up Interruptions\n");
    gdtSetup();
    idtSetup();
    irqSetup();
    pitSetup(HZ);

    if (magic == MULTIBOOT_BOOTLOADER_MAGIC) { // Get loaded by Grub with mutliboot version 1
        multiboot_info_t *mbi = (multiboot_info_t *)addr;
        /* Are mem_* valid? */
        if (CHECK_FLAG(mbi->flags, 0)) {
            printf("mem_lower = %dKiB mem_upper %dKiB\n", mbi->mem_lower, mbi->mem_upper);
            upper_mem = mbi->mem_upper;
        }

        /* Is boot_device valid? */
        if (CHECK_FLAG(mbi->flags, 1)) {
            printf("boot_device = %d\n", mbi->boot_device);
        }

        /* Is the command line passed? */
        if (CHECK_FLAG(mbi->flags, 2)) {
            printf("cmdline = %s\n", (char *)mbi->cmdline);
        }

        if (CHECK_FLAG(mbi->flags, 3)) {
            multiboot_module_t *mod;
            uint32_t i;

            printf("mods_count = %d, mods_addr = 0x%x\n", (int)mbi->mods_count,
                   (int)mbi->mods_addr);
            for (i = 0, mod = (multiboot_module_t *)mbi->mods_addr; i < mbi->mods_count;
                 i++, mod++)
                printf(" mod_start = 0x%x, mod_end = 0x%x, cmdline = %s\n",
                       (unsigned)mod->mod_start, (unsigned)mod->mod_end, (char *)mod->cmdline);
        }

        if (CHECK_FLAG(mbi->flags, 6)) {
            struct multiboot_mmap_entry *mmap = (struct multiboot_mmap_entry *)mbi->mmap_addr;
            uint size = mbi->mmap_length / sizeof(struct multiboot_mmap_entry);
            pr_devel("mmap buffer at %d size %d %d\n", mbi->mmap_addr, mbi->mmap_length,
                     sizeof(multiboot_memory_map_t));
            for (uint i = 0; i < size; i++) {
                printf("base_addr 0x%x 0x%x, length = 0x%x 0x%x, type = 0x%x\n",
                       (unsigned)(mmap[i].addr >> 32), (unsigned)(mmap[i].addr & 0xffffffff),
                       (unsigned)(mmap[i].len >> 32), (unsigned)(mmap[i].len & 0xffffffff),
                       (uint32_t)mmap[i].type);
            }
        }
    }

    if (upper_mem == 0) {
        printf("Cannot get upper phy mem bound. Using default value 32MB\n");
        upper_mem = 32 * 1024;
    }

    printf("Setting up Pagination\n");
    paddr_t lastUserByMem;
    memSetup(upper_mem, &lastUserByMem);
#ifdef RUN_TEST
    testPhymem();
#endif
    pagingSetup(lastUserByMem);

    printf("Setting up IRQ handlers\n");
    irqSetRoutine(IRQ_KEYBOARD, keyboard_handler);

    printf("Enabling HW interrupts\n");
    // Enabling the HW interrupts
    exceptionSetup();
    asm volatile("sti\n");

    printf("Setting up Serial link (115200)\n");
    serialSetup(115200);
    allocSetup();
    kthreadSetup(_stack_bottom, (_stack_top - _stack_bottom + 1));
    kthreadCreate("idle ", idleThread, NULL);
    irqSetRoutine(IRQ_TIMER, pit_handler);
#ifdef RUN_TEST
    run_test();
#endif
    printf("\nSystem init done\n");

    // There is no real caller behind this point
    // So finish this by ourself
    kthreadExit();
}
